use futures::TryStreamExt;
use nix::mount::MsFlags;
use serde::{Deserialize, Serialize};
use std::fs;
use std::net::Ipv4Addr;
use std::{env, io, io::Write, iter::Iterator, os::unix, path::Path, process};
// inspirado en https://github.com/superfly/init-snapshot/blob/public/src/bin/init/main.rs#L230

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "PascalCase")]
struct Config {
    env: Vec<String>,
    cmd: Option<Vec<String>>,
    entrypoint: Option<Vec<String>>,
}

#[derive(Serialize, Deserialize, Debug)]
struct ImageConfig {
    architecture: String,
    config: Config,
}

#[tokio::main]
async fn main() -> Result<(), io::Error> {
    let new_root = mount_new_root()?;
    // TODO: move to after chroot after we start mounting /proc (uses /proc/cmdline)
    setup_net_interface()
        .await
        .expect("couldn't set up network interace");
    let config_file = Path::new(&new_root).join(".fireactions-image-config.json");
    let config_text = fs::read(config_file)?;
    let config: ImageConfig = serde_json::from_slice(&config_text).unwrap();
    println!("{:#?}", config);

    unix::fs::chroot(new_root)?;
    env::set_current_dir("/")?;

    // https://github.com/opencontainers/image-spec/blob/b5ec432b1c946c09e1568b18ef70b654a93739f6/conversion.md#verbatim-fields
    let args = match (config.config.cmd, config.config.entrypoint) {
        (None, Some(args)) => args,
        (Some(args), None) => args,
        (None, None) => panic!("Invalid config"),
        (Some(cmd), Some(entrypoint)) => {
            let mut args = entrypoint.clone();
            args.append(&mut cmd.clone());
            args
        }
    };

    // TODO: setup hostname
    // TODO: setup /etc/hosts
    populate_resolvconf()?;

    let mut child = process::Command::new(&args[0])
        .args(&args[1..])
        .env_clear()
        .envs(config.config.env.iter().map(|s| s.split_once('=').unwrap()))
        .spawn()?;
    let status = child.wait()?;
    if !status.success() {
        println!("{}", status);
    }
    nix::sys::reboot::reboot(nix::sys::reboot::RebootMode::RB_AUTOBOOT)
        .map(|_| {})
        .unwrap();
    Ok(())
}

// async fn get_link(handle: &rtnetlink::Handle, name: &'static str) -> _ {
//     handle
//         .link()
//         .get()
//         .match_name(name.to_string())
//         .execute()
//         .try_next()
//         .await
//         .expect("no lo link found")
// }

async fn setup_net_interface() -> Result<(), rtnetlink::Error> {
    let (connection, handle, _) = rtnetlink::new_connection().unwrap();
    tokio::spawn(connection);

    println!("netlink: getting lo link");
    let lo = handle
        .link()
        .get()
        .match_name("lo".to_string())
        .execute()
        .try_next()
        .await?
        .expect("no lo link found");

    println!("netlink: setting lo link \"up\"");
    handle.link().set(lo.header.index).up().execute().await?;

    let cmdline = fs::read("/proc/cmdline").expect("couldn't read /proc/cmdline");
    let (guest_addr, host_addr) = cmdline
        .split(|c| *c == b' ')
        .map(|keyval| std::str::from_utf8(keyval).unwrap().split_once("="))
        .filter_map(|o| o)
        .find(|(key, _)| *key == "fireactions_ip")
        .expect("didn't find fireactions_ip cmdline")
        .1
        .split_once(":")
        .unwrap();

    println!("netlink: getting eth0 link");
    let eth0 = handle
        .link()
        .get()
        .match_name("eth0".to_string())
        .execute()
        .try_next()
        .await?
        .expect("no eth0 link found");

    println!("netlink: setting eth0 link \"up\"");
    handle
        .link()
        .set(eth0.header.index)
        .up()
        .mtu(1420)
        .execute()
        .await?;

    // XXX: la impl en init-snapshot es sketchy. pareciera que lo único que hace
    // es desactivar el chequeo de checksum de paquetes, osea una mejora de perf.
    // let _ = ethtool_set("eth0", ETHTOOL_SRXCSUM, 0);
    // ethtool_set("eth0", ETHTOOL_STXCSUM, 0)?;

    println!("netlink: adding ip");
    let address = handle.address();
    address
        .add(eth0.header.index, guest_addr.parse().unwrap(), 24)
        .execute()
        .await?;

    println!("netlink: adding route");
    let route = handle.route();
    route
        .add()
        .v4()
        .gateway(host_addr.parse().unwrap())
        .execute()
        .await?;
    // TODO: ipv6

    Ok(())
}

fn populate_resolvconf() -> Result<(), io::Error> {
    let mut file = fs::OpenOptions::new()
        .create(true)
        .write(true)
        .truncate(true)
        .open("/etc/resolv.conf")?;
    // XXX: usar otro DNS?
    write!(&mut file, "nameserver\t{}\n", "1.1.1.1")
}

fn mount_new_root() -> Result<&'static str, io::Error> {
    // process::Command::new("modprobe")
    //     .args(&["overlay"])
    //     .spawn()?
    //     .wait()?;

    let image_root = "/image_src";
    fs::create_dir_all(image_root)?;
    nix::mount::mount::<_, _, _, [u8]>(
        Some("/dev/vdb"),
        image_root,
        Some("squashfs"),
        MsFlags::MS_RDONLY,
        None,
    )
    .unwrap();

    let tmp_rw = "/tmp_rw";
    fs::create_dir_all(tmp_rw)?;
    nix::mount::mount::<_, _, _, [u8]>(
        Some("overlaytemp"),
        tmp_rw,
        Some("tmpfs"),
        MsFlags::empty(),
        None,
    )
    .unwrap();
    let tmp_upper = "/tmp_rw/upper";
    fs::create_dir_all(tmp_upper)?;
    let tmp_work = "/tmp_rw/work";
    fs::create_dir_all(tmp_work)?;

    let new_root = "/new_root";
    fs::create_dir_all(new_root)?;
    let overlay_opts = format!(
        "lowerdir={},upperdir={},workdir={}",
        image_root, tmp_upper, tmp_work
    );
    nix::mount::mount::<_, _, _, [u8]>(
        Some("overlay-root"),
        new_root,
        Some("overlay"),
        MsFlags::empty(),
        Some(overlay_opts.as_bytes()),
    )
    .unwrap();

    Ok(new_root)
}
